//================================================================
//
//	Smartronix Proprietary Information
//
//
//	File Name:	iic_receive_model.v
//	Author:	R. Jolly
//	Date:		October 14, 2008
//	Revision:	A
//
//
//	Development:	Environment Control Module UAV
//	Application:	Altera Stratix II
//	Description:	This file contains Verilog code that is designed
//				to receive and decode I2C serial streams. 
//
//
//
//
//	Revision History:
//
//	DATE		VERSION		DETAILS
//	10/14/08		A		Created by R Jolly
//
//
//
//
// Arguments:   out_data  is a 16-bit data port which is used to output
//                        the receive data.
//              out_valid is a positive pulse which indicates that the
//                        data on "data_out" is valid.
//              active    indicates whether the data stream is active.
//                        The data stream becomes active when comm_data
//                        changes from high to low while comm_clk is
//                        high.  The data stream becomes inactive if
//                        comm_data changes from low to high while
//                        comm_clk is high.
//              comm_data is the input data stream.  Data is valid from
//                        "delay time" after the negative edge of 
//                        comm_clk to the following negative edge of
//                        comm_clk.  The signal comm_data should not
//                        change while comm_clk is high except for
//                        activation and deactivation of the data
//                        stream.
//              comm_clk  is the data clock which defines the timing on
//                        comm_data.
//
//
//
//
//
//
//
//==================================================================

module iic_receive_model(
		clk,
		reset,

		iic_data,
		iic_clock,
		
		receive_start,

		data_out,
		valid_out,
		active
		);

   parameter	log_width	=	3;


   input                        clk;
   input                        reset;

   output [(1<<log_width)-1:0]	data_out;
   output                       valid_out;
   output                       active;

   input                        iic_data;
   output                       iic_clock;

   input wire                   receive_start;

   reg  [(1<<log_width)-1:0]	data_out;
   reg                          valid_out;
   reg                          receive_start_reg;
   reg                          active;

   reg                          last_data;
   reg                          last_clk;
   reg                          lastlast_data;
   reg                          lastlast_clock;
   wire                         sample_data;
   reg  [(1<<log_width)-1:0]	shift_reg;
   //reg		[15:0]			shift_reg;
   
   reg  [log_width-1:0]         bit_counter;
   reg  [4:0]                   out_phase;

   reg                          z1_iic_data;
   reg                          z2_iic_data;
   reg                          z3_iic_data;

   reg                          iic_clock;
   reg                          z1_iic_clock;
   reg                          z2_iic_clock;
   reg                          z3_iic_clock;

//--------------------------------------------
//
//

   assign    sample_data = (iic_clock & active ) ? 1'b1: 1'b0;
//
// Input signals to this function are asynchronous to the 
// 100 MHz clock in the FPGA. This change is an attempt
// at solving the bit-slip problem w/o significant
// effort. The assumption here is that the logic is correct,
// but that due to the asynchronous nature of the input signals,
// the function suffers from meta-stability issues, which
// cause occasional glitches in the logic. The following code
// section will ensure that the logic sees no setup or hold violations.


	always @(posedge clk)
	begin
		z1_iic_data = iic_data;
		z2_iic_data = z1_iic_data;
		z3_iic_data = z2_iic_data;
	end


//Main decode loop

	always @(posedge clk or negedge reset)
	begin
    if( !reset )
    begin
        iic_clock <= 1'b0;
        bit_counter <= 9'h0;
        data_out <= 8'h0;
        valid_out <= 1'b0;
        active <= 1'b0;
        shift_reg <= 16'h0000;
        out_phase <= 5'b00000;
        valid_out <= 1'b0;
        last_data <= 1'b1;
        last_clk <= 1'b1;
        lastlast_data <= 1'b1;
        lastlast_clock <= 1'b1;
        //sample_data <= 1'b0;
        receive_start_reg <= 1'b0;
        active <= 1'b0;
    end
	else
	begin
	
	    //Active Generation
	    if(receive_start & !receive_start_reg)
        begin
		    receive_start_reg <= 1'b1;
			active <= 1'b1;
			bit_counter <= (1<<log_width);
			out_phase <= 5'b10001;
		end
		else if(receive_start_reg & (out_phase == 5'b00000))
		begin
		    receive_start_reg <= 1'b0;
			active <= 1'b0;
		end
		
		//Clock Generation
		if(active)
		    iic_clock <= ~iic_clock;


		if ( sample_data )
		begin
			bit_counter <= bit_counter - 4'b0001;
			//sample_data <= 1'b0;
			//if ( bit_counter == 0 )
			//	out_phase <= 5'b00111;
			//else
				shift_reg	<= { shift_reg[(1<<log_width)-2:0],	z3_iic_data };
		end 
		//else if (z3_iic_clock & !last_clk & active )
		//else if (iic_clock & active )
		//begin
		//	sample_data <= 1'b1;
		//end


		case ( out_phase )
			5'b00001:	data_out <= shift_reg;
			5'b00010:	valid_out <= 1'b1;
			5'b00000:	valid_out <= 1'b0;
		endcase

		if ( out_phase != 5'b00000)
			out_phase <= out_phase - 5'b00001;


		last_data <= z3_iic_data;
		lastlast_data <= last_data;
	end
	end
endmodule


