//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2014
//#
//#
//# File Name:  serial_iic_comms_model.v
//# Author:     Earth People Technology
//# Date:       December 13, 2014
//# Revision:   A
//#
//# Development: EPT I2C Slave Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			12/13/14 	1			Created
//#
//#									
//#
//#######################################################################

module serial_iic_comms_model(
		CLK,
		RESET,

		SERIAL_CLK,

		IIC_DATA,
		IIC_CLOCK,

		SERIAL_IIC_TRANSMIT_DATA_IN,
		SERIAL_IIC_TRANSMIT_ENA_OUT,
		SERIAL_IIC_TRANSMIT_RDY_IN,
		SERIAL_IIC_TRANSMIT_WORD_COUNT_IN,
		SERIAL_IIC_TRANSMIT_ADDRESS_IN,
		
		SERIAL_I2C_RECEIVE_REQUEST,

		SERIAL_IIC_RECEIVE_DATA_OUT,
		SERIAL_IIC_RECEIVE_ENA_IN,
		SERIAL_IIC_RECEIVE_RDY_OUT
		);

//==================================================================
// Parameter Declarations
//==================================================================
    parameter IDLE                                = 0,
              I2C_START                           = 1,
              RECEIVE_WORD                        = 2,
              STORE_DATA_WORD                     = 3,
              CHECK_FOR_LAST_WORD                 = 4,
              READ_WORD_DELAY                     = 5,
              RECEIVE_CHECK_ACK                   = 6,
              WRITE_TO_I2C_BUS                    = 7,
              RECEIVE_TRANSMIT_MUX                = 8,
              CHECK_FOR_TRANSMIT_WORD             = 9,
              CHECK_FOR_TRANSMIT_WORD_DELAY       = 10,
              ADDRESS_WORD                        = 11,
              ADDRESS_WORD_DELAY                  = 12,
              WRITE_OUTGOING_MESSAGE              = 13,
              WAIT_FOR_WRITE_CYCLE_COMPLETE       = 14,
              WRITE_CYCLE_COMPLETE                = 15,
              TRANSMIT_CHECK_ACK                  = 16,
              I2C_STOP                            = 17;


//==================================================================
// IO Port Declarations
//==================================================================
	input						CLK;
	input						RESET;

	input						SERIAL_CLK;

	inout						IIC_DATA;
	output						IIC_CLOCK;

	input   [7:0]               SERIAL_IIC_TRANSMIT_DATA_IN;
	output                      SERIAL_IIC_TRANSMIT_ENA_OUT;
	input                       SERIAL_IIC_TRANSMIT_RDY_IN;
	input   [7:0]               SERIAL_IIC_TRANSMIT_WORD_COUNT_IN;
	input   [7:0]               SERIAL_IIC_TRANSMIT_ADDRESS_IN;
	
	input                       SERIAL_I2C_RECEIVE_REQUEST;

	output  [7:0]               SERIAL_IIC_RECEIVE_DATA_OUT;
	input                       SERIAL_IIC_RECEIVE_ENA_IN;
	output                      SERIAL_IIC_RECEIVE_RDY_OUT;


//==================================================================
// Register Declarations
//==================================================================
    wire	[7:0]				SERIAL_IIC_RECEIVE_DATA_OUT;
    reg                         SERIAL_IIC_TRANSMIT_ENA_OUT;
    reg                         SERIAL_IIC_RECEIVE_RDY_OUT;

    reg                         serial_iic_ena_out_reg;
  
  	// Serial stream signals
    wire						IIC_DATA;
    wire						IIC_CLOCK;

	// State Machine Registers
    reg	[17:0]                  state;
    reg	[17:0]                  next;

	// Receive FIFO Storage
    reg                         receive_fifo_rden;
    reg                         receive_fifo_rden_reg;
    wire	[15:0]				receive_fifo_din;
    reg                         receive_fifo_wren;
    reg                         receive_fifo_wren_reg;
    wire	[15:0]				receive_fifo_dout;
    wire						receive_fifo_full;
    wire						receive_fifo_empty;
    wire						receive_data_count;
    wire                        receive_fifo_wr_ack;
    wire                        receive_fifo_afull;
    wire                        receive_fifo_aempty;
    wire                        receive_fifo_flush;

	// Transmit FIFO Storage
    reg                         transmit_fifo_rden;
    wire	[15:0]				transmit_fifo_din;
    reg                         transmit_fifo_wren;
    wire	[15:0]				transmit_fifo_dout;
    wire						transmit_fifo_full;
    wire						transmit_fifo_empty;
    wire						transmit_data_count;
    wire                        transmit_fifo_wr_ack;
    wire                        transmit_fifo_afull;
    wire                        transmit_fifo_aempty;
    wire                        transmit_fifo_flush;

	// Read Cycle Status Messages
	reg   [1:0]                 serial_iic_rdy_out_reg;

	// FIFO Status Messages
//	reg                         serial_iic_rdy_out;

	wire						iic_data_to_receive;

	wire	[7:0]               receive_data_out;
	wire						receive_valid_out;
	wire						receive_active;

	// IIC Transmit unit Registers
	wire						iic_data_from_transmit;
	wire						iic_clock_from_transmit;
	reg   [7:0]                 transmit_data_in;
	reg                         transmit_data_valid;
	reg                         transmit_active;
	reg                         transmit_active_reg;
	wire                        transmit_done;
	wire                        transmit_start;
	reg                         transmit_data_valid_reg;
	reg                         transmit_data_complete;

	//Start I2C state machine
 	reg                         write_to_iic;			
	reg  [1:0]                  write_to_iic_state ;
	reg  [7:0]                  read_word_out_count;
	
  	// Receive registers
	wire                        iic_clock_from_receive;
	reg                         receive_start;

	// Serial IIC DATA and CLOCK multiplexor
	reg                         transmission_active;
	wire                        reception_active;

	// Incoming word counter
	reg	[7:0]					incoming_word_count ;
	
	//I2C Read/Write Bit
	reg                         i2c_readwrite_bit;
	reg                         i2c_readwrite_bit_reg; 
	
	//I2C Ack Unit registers
    wire                        ack_grant_start;
    wire                        ack_req_start;
    wire                        i2c_ack_receive;
    wire [7:0]                  i2c_ack_data_out;
    wire                        i2c_ack_transmit;
    wire                        i2c_ack_done;
    wire                        i2c_ack_clock;
	wire                        i2c_ack_rcv_active;
	wire                        i2c_ack_send_active;
	
	//I2C Start Registers
	reg                         i2c_start_active;
	reg  [1:0]                  i2c_start_active_reg;
	reg                         i2c_start_data;
	reg                         i2c_start_clock;
	
	//I2C Stop Registers
	reg                         i2c_stop_active;
	reg  [1:0]                  i2c_stop_active_reg;
	reg                         i2c_stop_data;
	reg                         i2c_stop_clock;

	

//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

	assign	IIC_DATA = i2c_start_active ? i2c_start_data : (i2c_ack_rcv_active ?  1'bz : (transmission_active ? iic_data_from_transmit : (reception_active ? 1'bz : (i2c_ack_send_active ? i2c_ack_receive : (i2c_stop_active ? i2c_stop_data : 1'b1)))));
	assign	IIC_CLOCK = i2c_start_active ? i2c_start_clock : (transmission_active ? iic_clock_from_transmit : (receive_active ? iic_clock_from_receive : (i2c_stop_active ? i2c_stop_clock : 1'b1)));


	assign	iic_data_to_receive = !transmission_active ? IIC_DATA : 1'bz;

	assign	receive_fifo_din = receive_data_out;

	assign	SERIAL_IIC_RECEIVE_DATA_OUT = receive_fifo_dout;
 	
	assign	transmit_fifo_din = SERIAL_IIC_TRANSMIT_DATA_IN;
	
	//I2C Receive Registers
	assign	reception_active = receive_active & !receive_valid_out;

	//I2C Ack Unit registers
	assign	ack_grant_start = receive_active;
	assign	ack_req_start = transmit_active;
	//assign	i2c_ack_transmit = iic_data_from_transmit;
	assign	i2c_ack_transmit = IIC_DATA;
	assign	i2c_ack_clock = iic_clock_from_transmit | iic_clock_from_receive;
	

//=================================================================================
//  I2C Start: Send I2C start across bus
//=================================================================================

	always @( negedge SERIAL_CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
            i2c_start_active <= 1'b0;
			i2c_start_active_reg <= 2'b00;
            i2c_start_data <= 1'b1;
            i2c_start_clock <= 1'b1;
		end
  		else
    		begin
        		if(state[I2C_START] & (i2c_start_active_reg == 2'b00))
          		begin
              		i2c_start_active <= 1'b1;
			        i2c_start_active_reg <= 2'b01;
					i2c_start_data <= 1'b0;
					i2c_start_clock <= 1'b1;
          		end
       		    else if(state[I2C_START] & (i2c_start_active_reg == 2'b01))
          		begin
              		i2c_start_active <= 1'b1;
			        i2c_start_active_reg <= 2'b10;
					i2c_start_data <= 1'b0;
					i2c_start_clock <= 1'b0;
          		end
       		    else if((i2c_start_active_reg == 2'b10))
          		begin
              		i2c_start_active <= 1'b0;
			        i2c_start_active_reg <= 2'b00;
					i2c_start_data <= 1'b1;
					i2c_start_clock <= 1'b1;
          		end
    		end
  	end

//=================================================================================
//  I2C Stop: Send I2C stop across bus
//=================================================================================

	always @( negedge SERIAL_CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
            i2c_stop_active <= 1'b0;
			i2c_stop_active_reg <= 2'b00;
            i2c_stop_data <= 1'b0;
            i2c_stop_clock <= 1'b0;
		end
  		else
    		begin
        		if(transmission_active & (i2c_stop_active_reg == 2'b00))
          		begin
              		i2c_stop_active <= 1'b1;
			        i2c_stop_active_reg <= 2'b01;
					i2c_stop_data <= 1'b0;
					i2c_stop_clock <= 1'b0;
          		end
				//else if(state[WRITE_CYCLE_COMPLETE] & (i2c_stop_active_reg == 2'b01))
				else if(state[I2C_STOP] & (i2c_stop_active_reg == 2'b01))
          		begin
              		i2c_stop_active <= 1'b1;
			        i2c_stop_active_reg <= 2'b10;
					i2c_stop_data <= 1'b0;
					i2c_stop_clock <= 1'b0;
          		end
       		    //else if(state[WRITE_CYCLE_COMPLETE] & (i2c_stop_active_reg == 2'b10))
       		    else if(state[I2C_STOP] & (i2c_stop_active_reg == 2'b10))
          		begin
              		i2c_stop_active <= 1'b1;
			        i2c_stop_active_reg <= 2'b11;
					i2c_stop_data <= 1'b0;
					i2c_stop_clock <= 1'b1;
          		end
       		    else if((i2c_stop_active_reg == 2'b11))
          		begin
              		i2c_stop_active <= 1'b0;
			        i2c_stop_active_reg <= 2'b00;
					i2c_stop_data <= 1'b1;
					i2c_stop_clock <= 1'b1;
          		end
    		end
  	end

//=================================================================================
//  Read/Write Bit: Set the Read/Write Bit
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
            i2c_readwrite_bit <= 1'b0;
			i2c_readwrite_bit_reg <= 1'b0;
		end
  		else
    		begin
        		if(SERIAL_I2C_RECEIVE_REQUEST & !i2c_readwrite_bit_reg)
          		begin
              		i2c_readwrite_bit <= 1'b1;
			        i2c_readwrite_bit_reg <= 1'b1;
          		end
				else if(state[I2C_STOP] & i2c_readwrite_bit_reg)
          		begin
              		i2c_readwrite_bit <= 1'b0;
			        i2c_readwrite_bit_reg <= 1'b0;
          		end
            end
  	end
	
//=================================================================================
//  Write Cycle: Set the transmit_active
//=================================================================================

    always @( negedge CLK or negedge RESET )
    begin
        if ( !RESET )
		begin
          transmit_active <= 1'b0;
		  transmit_active_reg <= 1'b0;
        end
        else
        begin
           if(state[ADDRESS_WORD] | (transmit_start & state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & !transmit_active_reg))
           begin
              transmit_active <= 1'b1;
		      transmit_active_reg <= 1'b0;
           end
           else if((transmit_start & state[ADDRESS_WORD_DELAY]))
		   begin
              transmit_active <= 1'b0;
 		      transmit_active_reg <= 1'b0;
           end
          else if(!transmit_start & state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & !transmit_active_reg)
		  begin
              transmit_active <= 1'b1;
		      transmit_active_reg <= 1'b1;
           end
          else if(transmit_start & state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & transmit_active_reg)
		  begin
              transmit_active <= 1'b0;
		      transmit_active_reg <= 1'b1;
           end
          else if(transmit_start & state[WRITE_CYCLE_COMPLETE] & transmit_active_reg)
		  begin
              transmit_active <= 1'b0;
		      transmit_active_reg <= 1'b0;
           end
        end
    end

//=================================================================================
//  Write Cycle: Select Address or Data to Transmit
//=================================================================================

    always @( negedge CLK or negedge RESET )
    begin
        if ( !RESET )
		begin
          transmit_data_in <= 8'h0;
    end
    else
    begin
        if(state[ADDRESS_WORD])
          begin
              transmit_data_in <= {SERIAL_IIC_TRANSMIT_ADDRESS_IN[7:1], i2c_readwrite_bit};
          end
        else if(state[WRITE_OUTGOING_MESSAGE])
            transmit_data_in <= transmit_fifo_dout;
    end
    end

//=================================================================================
//  Write Cycle: Determine Receive cycle
//=================================================================================

    always @( negedge CLK or negedge RESET )
    begin
        if ( !RESET )
		begin
          read_word_out_count <= 8'h0;
    end
    else
    begin
        //if(SERIAL_IIC_TRANSMIT_RDY_IN)
        if(SERIAL_I2C_RECEIVE_REQUEST)
          begin
              read_word_out_count <= SERIAL_IIC_TRANSMIT_WORD_COUNT_IN;
          end
        //if (state[I2C_STOP])		  
        if (state[CHECK_FOR_LAST_WORD])
            read_word_out_count <= 8'h0;
    end
    end
	
//=================================================================================
//	Write Cycle: Monitor the iic_serial_rdy_in signal for Write Cycle Command
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			write_to_iic <= 1'b0;			
			write_to_iic_state <= 2'b00;
		end
		else
		begin
			case ( write_to_iic_state )
			2'b00:
			begin
			    if(!transmit_fifo_empty)
				begin
				   write_to_iic <= 1'b1;
				   write_to_iic_state <= 2'b01;
				end
                else if(read_word_out_count > 0)
                begin				
				   write_to_iic <= 1'b1;
				   write_to_iic_state <= 2'b10;
				end
		    end
			2'b01:
			begin
				if(transmit_fifo_empty)
				begin
				   write_to_iic <= 1'b0;
				   write_to_iic_state <= 2'b00;
				end
		    end
			2'b10:
			begin
				if(state[RECEIVE_WORD])
				begin
				   write_to_iic <= 1'b0;
				   write_to_iic_state <= 2'b11;
				end
		    end
			2'b11:
			begin
			    if(incoming_word_count == 0)
				begin
				   write_to_iic <= 1'b0;
				   write_to_iic_state <= 2'b00;
				end
		    end
			default: 
			begin
				   write_to_iic <= 1'b0;
				   write_to_iic_state <= 2'b00;
		    end
			endcase
		
		end
end

//=================================================================================
//  Write Cycle: transmission_active register
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			transmission_active <= 1'b0;
		end
  		else
    		begin
        		if(state[WRITE_TO_I2C_BUS])
          		begin
              		transmission_active <= 1'b1;
          		end
       		 else if(state[RECEIVE_WORD] | state[WRITE_CYCLE_COMPLETE])
            		 transmission_active <= 1'b0;
    		end
  	end



//=================================================================================
//	Write Cycle: Send word from transmit FIFO to transmit unit
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			transmit_data_valid <= 1'b0;
			transmit_data_valid_reg <= 1'b0;
			transmit_data_complete <= 1'b0;
			transmit_fifo_rden <= 1'b0;
			transmit_data_complete <= 1'b0;
		end
		else
		begin
			if(state[ADDRESS_WORD])
			begin
					transmit_data_valid <= 1'b1;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 1'b0;
					transmit_data_complete <= 1'b0;
			end
			if(state[ADDRESS_WORD_DELAY])
			begin
					transmit_data_valid <= 1'b0;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 1'b0;
					transmit_data_complete <= 1'b0;
			end
			if(state[CHECK_FOR_TRANSMIT_WORD])
			begin
					transmit_data_valid <= 1'b0;
					transmit_fifo_rden <= 1'b1;
					transmit_data_valid_reg <= 1'b0;
					transmit_data_complete <= 1'b0;
			end	
			else if( state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & transmit_start & !transmit_data_valid_reg)
			begin
					transmit_data_valid <= 1'b1;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 1'b0;
			end
			else if( state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & !transmit_start & !transmit_data_valid_reg)
			begin
				    transmit_data_valid <= 1'b0;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 1'b1;
					transmit_data_complete <= 1'b1;
			end
			else if( state[IDLE] )
			begin
				transmit_data_valid <= 1'b0;
				transmit_fifo_rden <= 1'b0;
				transmit_data_valid_reg <= 1'b0;
				transmit_data_complete <= 1'b0;
			end
		end
end

//=================================================================================
//	Write Cycle: Store incoming word into the transmit FIFO
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
		        transmit_fifo_wren <= 1'b0;
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
				serial_iic_ena_out_reg <= 1'b0;
		end
		else 
		begin
			if(SERIAL_IIC_TRANSMIT_RDY_IN  & !serial_iic_ena_out_reg )
			begin
		        transmit_fifo_wren <= 1'b1;
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b1;
				serial_iic_ena_out_reg <= 1'b1;
			end
			else if(!SERIAL_IIC_TRANSMIT_RDY_IN  & serial_iic_ena_out_reg )
			begin
		        transmit_fifo_wren <= 1'b0;
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
				serial_iic_ena_out_reg <= 1'b1;
			end
			else if(!SERIAL_IIC_TRANSMIT_ENA_OUT & SERIAL_IIC_TRANSMIT_RDY_IN  )
			begin
		        transmit_fifo_wren <= 1'b0;
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
				serial_iic_ena_out_reg <= 1'b0;
			end
		end
	end

//=================================================================================
//  Receive: receive_start register
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_start <= 1'b0;
		end
		else
		begin
			if(state[RECEIVE_WORD])
				receive_start <= 1'b1;
			else if(receive_valid_out)
				receive_start <=1'b0;
		end
	end


//=================================================================================
//  Receive: Assert the SERIAL_IIC_RECEIVE_RDY_OUT register when receive FIFO has data
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			SERIAL_IIC_RECEIVE_RDY_OUT <= 1'b0;
			serial_iic_rdy_out_reg <= 2'b00;
		end
		else
		begin
			if(!receive_fifo_empty & (serial_iic_rdy_out_reg == 2'b00))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <= 1'b1;
				serial_iic_rdy_out_reg <= 2'b01;
			end
			else if(SERIAL_IIC_RECEIVE_ENA_IN & (serial_iic_rdy_out_reg == 2'b01))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <=1'b0;
				serial_iic_rdy_out_reg <= 2'b10;
			end
			else if (!SERIAL_IIC_RECEIVE_ENA_IN & (serial_iic_rdy_out_reg == 2'b10))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <=1'b0;
				serial_iic_rdy_out_reg <= 2'b00;
			end
		end
	end

/*
	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			serial_iic_rdy_out <= 1'b0;
		end
		else
		begin
			if( !receive_fifo_empty )
				serial_iic_rdy_out <= 1'b1;
			else if ( receive_fifo_empty )
				serial_iic_rdy_out <=1'b0;
		end
	end
*/

//=================================================================================
//  Receive: Read word from receive FIFO to xxxx_feedback.v when fifo_rden is asserted
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_fifo_rden <= 1'b0;
			receive_fifo_rden_reg <= 1'b0;
		end
		else
		begin
			if( SERIAL_IIC_RECEIVE_ENA_IN & !receive_fifo_rden_reg )
			begin
				receive_fifo_rden <= 1'b1;
				receive_fifo_rden_reg <= 1'b1;
			end
			else if( SERIAL_IIC_RECEIVE_ENA_IN & receive_fifo_rden_reg )
			begin
				receive_fifo_rden <= 1'b0;
				receive_fifo_rden_reg <= 1'b1;
			end
			else if( !SERIAL_IIC_RECEIVE_ENA_IN & receive_fifo_rden_reg  )
			begin
				receive_fifo_rden <= 1'b0;
				receive_fifo_rden_reg <= 1'b0;
			end
		end
	end

//=================================================================================
//  Receive: Create the incoming_word_count for the receive unit 
//=================================================================================
	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			incoming_word_count <= 8'h0;
		end
		else
		begin
		    if(state[RECEIVE_TRANSMIT_MUX] & read_word_out_count)
			    incoming_word_count <= read_word_out_count;
			//else if(state[STORE_DATA_WORD] & receive_fifo_wren)
			else if(state[STORE_DATA_WORD])
			begin
			   if(incoming_word_count > 8'h0)
                  incoming_word_count <= incoming_word_count - 1;
			end
			else if( state[IDLE] )
				incoming_word_count <= 8'h0;


		end
	end


//=================================================================================
//  Receive: Load valid word from receive unit into FIFO when valid_out is asserted
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_fifo_wren <= 1'b0;
			receive_fifo_wren_reg <= 1'b0;
		end
		else
		begin
			if(state[STORE_DATA_WORD] & !receive_fifo_wren_reg)
			begin
				receive_fifo_wren <= 1'b1;
				receive_fifo_wren_reg <= 1'b1;
			end
			else if(state[STORE_DATA_WORD] & receive_fifo_wren_reg)
			begin
				receive_fifo_wren <= 1'b0;
				receive_fifo_wren_reg <= 1'b1;
			end
			else if( state[CHECK_FOR_LAST_WORD])
			begin
				receive_fifo_wren <= 1'b0;
				receive_fifo_wren_reg <= 1'b0;
			end
		end
	end
		
//=================================================================================
//	Serial I2C Communications Finite State Machine
//=================================================================================

	always @(  posedge CLK or negedge RESET  )
	begin
		if( !RESET )
		begin
			state <= 18'h000;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( CLK or state or transmit_data_valid or transmit_start or 
          transmit_fifo_wren or transmit_fifo_rden or write_to_iic_state or 
          transmit_fifo_empty or receive_fifo_wren or receive_fifo_rden or 
          receive_fifo_empty or receive_active or receive_valid_out or 
		  i2c_ack_done or i2c_ack_rcv_active or iic_clock_from_receive)
	begin
		next = 18'h000;
		if (state[IDLE])
		begin
			if(write_to_iic)
				next[I2C_START] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if (state[I2C_START])
		begin
			if(receive_active)
				next[STORE_DATA_WORD] = 1'b1;
			else if (i2c_start_active_reg == 2'b10)
				next[WRITE_TO_I2C_BUS] = 1'b1;
			else
				next[I2C_START] = 1'b1;
		end

		if (state[RECEIVE_WORD])
		begin
			if (i2c_ack_send_active)
				next[RECEIVE_CHECK_ACK] = 1'b1;
			else
				next[RECEIVE_WORD] = 1'b1;
		end

		if (state[STORE_DATA_WORD])
		begin
			if (i2c_ack_done)
				next[CHECK_FOR_LAST_WORD] = 1'b1;
			else
				next[STORE_DATA_WORD] = 1'b1;
		end

		if (state[CHECK_FOR_LAST_WORD])
		begin
			if ( incoming_word_count > 4'h0 )
				next[READ_WORD_DELAY] = 1'b1;
			else if(!iic_clock_from_receive)
				next[I2C_STOP] = 1'b1;
			else 
				next[CHECK_FOR_LAST_WORD] = 1'b1;
			
		end

		if (state[READ_WORD_DELAY])
		begin
			if (!receive_valid_out)
				next[RECEIVE_WORD] = 1'b1;
			else
				next[READ_WORD_DELAY] = 1'b1;
		end

		if (state[RECEIVE_CHECK_ACK])
		begin
			if (i2c_ack_done)
				//next[RECEIVE_WORD] = 1'b1;
				next[STORE_DATA_WORD] = 1'b1;
			else
				next[RECEIVE_CHECK_ACK] = 1'b1;
		end

		if (state[WRITE_TO_I2C_BUS])
		begin
			//if ( write_to_iic_bus_count >= 4'h2 )
				next[ADDRESS_WORD] = 1'b1;
			//else
			//	next[WRITE_TO_I2C_BUS] = 1'b1;
		end

		if (state[ADDRESS_WORD])
		begin
			if (!transmit_start)
				next[ADDRESS_WORD_DELAY] = 1'b1;
			else
				next[ADDRESS_WORD] = 1'b1;
		end

		if (state[ADDRESS_WORD_DELAY]) 
		begin
			if(i2c_ack_rcv_active)
				next[TRANSMIT_CHECK_ACK] = 1'b1;
			else
				next[ADDRESS_WORD_DELAY] = 1'b1;
		end

		if (state[RECEIVE_TRANSMIT_MUX])
		begin
			if(!transmit_fifo_empty)
				next[CHECK_FOR_TRANSMIT_WORD] = 1'b1;
			else if(read_word_out_count != 0)
				next[RECEIVE_WORD] = 1'b1;
			else
				next[I2C_STOP] = 1'b1;
		end
		
		if (state[CHECK_FOR_TRANSMIT_WORD])
		begin
			if(!transmit_fifo_empty)
				next[WRITE_OUTGOING_MESSAGE] = 1'b1;
			else
				next[I2C_STOP] = 1'b1;
		end

		if (state[WRITE_OUTGOING_MESSAGE])
		begin
			if(!transmit_done)
				next[WAIT_FOR_WRITE_CYCLE_COMPLETE] = 1'b1;
			else
				next[WRITE_OUTGOING_MESSAGE] = 1'b1;
		end


		if (state[WAIT_FOR_WRITE_CYCLE_COMPLETE])
		begin
			if (transmit_done)
				next[WRITE_CYCLE_COMPLETE] = 1'b1;
			else
				next[WAIT_FOR_WRITE_CYCLE_COMPLETE] = 1'b1;
		end

		if (state[WRITE_CYCLE_COMPLETE])
		begin
			if (!transmit_done)
				next[CHECK_FOR_TRANSMIT_WORD] = 1'b1;
			else
				next[WRITE_CYCLE_COMPLETE] = 1'b1;
		end

		if (state[TRANSMIT_CHECK_ACK]) 
		begin
			if(transmit_done & write_to_iic)
				next[RECEIVE_TRANSMIT_MUX] = 1'b1;
			else if(transmit_done & !write_to_iic)
				next[CHECK_FOR_TRANSMIT_WORD] = 1'b1;
			else
				next[TRANSMIT_CHECK_ACK] = 1'b1;
		end

		if ( state[I2C_STOP] )
		begin
			//if (!transmit_fifo_empty)		
			if(i2c_stop_active_reg == 2'b11)
				next[IDLE] = 1'b1;
			else
				next[I2C_STOP] = 1'b1;
		end


	end


//=================================================================
// For Simumlation, display the state names
//
//    parameter IDLE                                = 0,
//              I2C_START                           = 1,
//              RECEIVE_WORD                        = 2,
//              STORE_DATA_WORD                     = 3,
//              CHECK_FOR_LAST_WORD                 = 4,
//              READ_WORD_DELAY                     = 5,
//              RECEIVE_CHECK_ACK                   = 6,
//              WRITE_TO_I2C_BUS                    = 7,
//              RECEIVE_TRANSMIT_MUX                = 8,
//              CHECK_FOR_TRANSMIT_WORD             = 9,
//              CHECK_FOR_TRANSMIT_WORD_DELAY       = 10,
//              ADDRESS_WORD                        = 11,
//              ADDRESS_WORD_DELAY                  = 12,
//              WRITE_OUTGOING_MESSAGE              = 13,
//              WAIT_FOR_WRITE_CYCLE_COMPLETE       = 14,
//              WRITE_CYCLE_COMPLETE                = 15,
//              TRANSMIT_CHECK_ACK                  = 16,
//              I2C_STOP                            = 17;
//=================================================================

`ifdef SIM
always @( posedge CLK )
begin
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << I2C_START))
		state_name = "I2C_START";		
	if ( state == (1 << RECEIVE_WORD))
		state_name = "RECEIVE_WORD";		
	if ( state == (1 << STORE_DATA_WORD))
		state_name = "STORE_DATA_WORD";		
	if ( state == (1 << CHECK_FOR_LAST_WORD))
		state_name = "CHECK_FOR_LAST_WORD";
	if ( state == (1 << READ_WORD_DELAY))
		state_name = "READ_WORD_DELAY";
	if ( state == (1 << RECEIVE_CHECK_ACK))
		state_name = "RECEIVE_CHECK_ACK";
	if ( state == (1 << WRITE_TO_I2C_BUS))
		state_name = "WRITE_TO_I2C_BUS";
	if ( state == (1 << RECEIVE_TRANSMIT_MUX))
		state_name = "RECEIVE_TRANSMIT_MUX";
	if ( state == (1 << CHECK_FOR_TRANSMIT_WORD))
		state_name = "CHECK_FOR_TRANSMIT_WORD";
	if ( state == (1 << CHECK_FOR_TRANSMIT_WORD_DELAY))
		state_name = "CHECK_FOR_TRANSMIT_WORD_DELAY";
	if ( state == (1 << ADDRESS_WORD))
		state_name = "ADDRESS_WORD";
	if ( state == (1 << ADDRESS_WORD_DELAY))
		state_name = "ADDRESS_WORD_DELAY";
	if ( state == (1 << WRITE_OUTGOING_MESSAGE))
		state_name = "WRITE_OUTGOING_MESSAGE";
	if ( state == (1 << WAIT_FOR_WRITE_CYCLE_COMPLETE))
		state_name = "WAIT_FOR_WRITE_CYCLE_COMPLETE";
	if ( state == (1 << WRITE_CYCLE_COMPLETE))
		state_name = "WRITE_CYCLE_COMPLETE";
	if ( state == (1 << TRANSMIT_CHECK_ACK))
		state_name = "TRANSMIT_CHECK_ACK";
	if ( state == (1 << I2C_STOP))
		state_name = "I2C_STOP";
end
`endif		



//==============================================================================
//	Instantiate the IIC Transmit Interface
//==============================================================================
	iic_transmit_model		serial_transmit_inst
	(
		.clk                    (SERIAL_CLK),
		.reset                  (RESET),

		.iic_data               (iic_data_from_transmit),
		.iic_clock              (iic_clock_from_transmit),

		.holding_reg_empty      (transmit_start),
		.transmit_complete      (transmit_done),
		.data_in                (transmit_data_in),
		.data_valid             (transmit_data_valid),
		.active                 (transmit_active)
	);


//==============================================================================
//	Instantiate the IIC Receive Interface
//==============================================================================
	iic_receive_model		    serial_receive_inst
	(
		.clk                    (SERIAL_CLK),
		.reset                  (RESET),

		.iic_data				(iic_data_to_receive),
		.iic_clock				(iic_clock_from_receive),
		
		.receive_start          (receive_start),

		.data_out               (receive_data_out),
		.valid_out              (receive_valid_out),
		.active                 (receive_active)
	);

//==============================================================================
//	Instantiate the I2C Ack Interface
//==============================================================================
	i2c_ack_model               I2C_ACK_MODEL_INST
	(
      .CLK                      (CLK),
      .RST_N                    (RESET),
      
      .GNT_START                (ack_grant_start),
      .REQ_START                (ack_req_start),
      .DATA_IN                  (i2c_ack_transmit),
      .VALID_SEND               (i2c_ack_send_active),
	  .VALID_RECEIVE            (i2c_ack_rcv_active),
		

      .DOUT                     (i2c_ack_data_out),
      .DATA_OUT                 (i2c_ack_receive),
      .DONE                     (i2c_ack_done),
      .SCLK                     (i2c_ack_clock)
	  
	);
	
//-----------------------------------------------
// Instantiate FIFO for receiving incoming data 
//	from IIC bus fifo2048x8  module
//-----------------------------------------------
   
  sync_fifo #(
      .ADDR_WIDTH               (3),
	  .DATA_WIDTH               (8),
	  .DEPTH                    (8),
	  .AEMPTY                   (1),
	  .AFULL                    (6)
	  ) I2C_RECEIVE_FIFO
     (
	  .clk                     (CLK),
	  .reset_n                 (RESET),
	  .flush                   (receive_fifo_flush),
	  .write_data              (receive_fifo_din),	  
	  .read_req                (receive_fifo_rden),
	  .wdata_valid             (receive_fifo_wren),
	  .read_data               (receive_fifo_dout),
	  .rdata_valid             (),
	  .fifo_empty              (receive_fifo_empty),
	  .fifo_aempty             (receive_fifo_aempty),
	  .fifo_full               (receive_fifo_full),
	  .fifo_afull              (receive_fifo_afull),
	  .write_ack               (receive_fifo_wr_ack)
      );

/*   
   
    fifo_4x16_model             receive_FIFO
     (
      .DIN                      (receive_fifo_din),

      .WR_EN                    (receive_fifo_wren),
      .WR_CLK                   (CLK),

      .RD_EN                    (receive_fifo_rden),
      .RD_CLK                   (CLK),

      .AINIT                    (RESET),
      .DOUT                     (receive_fifo_dout),
      .FULL                     (receive_fifo_full),
      .EMPTY                    (receive_fifo_empty),

      .WR_ERR                   (receive_data_count)
      );
 */  
//-----------------------------------------------
// Instantiate FIFO for storing outgoing data 
//	to the IIC bus fifo2048x8  module
//-----------------------------------------------
   sync_fifo #(
      .ADDR_WIDTH               (3),
	  .DATA_WIDTH               (8),
	  .DEPTH                    (8),
	  .AEMPTY                   (1),
	  .AFULL                    (6)
	  ) I2C_TRANSMIT_FIFO
     (
	  .clk                     (CLK),
	  .reset_n                 (RESET),
	  .flush                   (transmit_fifo_flush),
	  .write_data              (transmit_fifo_din),	  
	  .read_req                (transmit_fifo_rden),
	  .wdata_valid             (transmit_fifo_wren),
	  .read_data               (transmit_fifo_dout),
	  .rdata_valid             (),
	  .fifo_empty              (transmit_fifo_empty),
	  .fifo_aempty             (transmit_fifo_aempty),
	  .fifo_full               (transmit_fifo_full),
	  .fifo_afull              (transmit_fifo_afull),
	  .write_ack               (transmit_fifo_wr_ack)
      );
  
/*   
   fifo_4x16_model              transmit_FIFO
     (
      .DIN                      (transmit_fifo_din),

      .WR_EN                    (transmit_fifo_wren),
      .WR_CLK                   (CLK),

      .RD_EN                    (transmit_fifo_rden),
      .RD_CLK                   (CLK),

      .AINIT                    (RESET),
      .DOUT                     (transmit_fifo_dout),
      .FULL                     (transmit_fifo_full),
      .EMPTY                    (transmit_fifo_empty),

      .WR_ERR                   (transmit_data_count)
      );
*/

endmodule
