//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2023
//#
//#
//# File Name:  EPT_5M57_AP_M4_Top.v
//# Author:     R. Jolly
//# Date:       January 14, 2023
//# Revision:   A
//#
//# Development: USB Test Tool Interface board 
//# Application: Altera MAX V CPLD
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. The MAX V receives its commands via
//#				 USB.
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			1/14/23 	A			Created			RJJ
//#                     			RJJ
//#
//#									
//#
//#######################################################################

`timescale 1ns/1ps



//************************************************************************
//* Module Declaration
//************************************************************************

module EPT_5M57_AP_M4_Top (

	
	input  wire                CLK_66MHZ,
	input  wire                RST,

	//UART COMMAND BUS
	output wire                UART_OUT, //To external 
    input wire                 UART_IN, //From external
  
	input wire   [7:0]         LB_XIOLA,       //XIOH -- J16
	input wire   [2:0]         LB_XIOLC,       //XIOH -- J16
	input wire   [7:0]         LB_XIOH,        //XIOL -- J16
	input wire   [7:0]         LB_COMM,        //COMM -- J10
	input wire   [7:0]         LB_AD,          //XIOH -- J8
	input wire   [7:0]         LB_IOH,         //XIOH -- J9
	
	//Transceiver Control Signals
	output wire                TR_DIR_1,
	output wire                TR_OE_1,
	
	output wire                TR_DIR_2,
	output wire                TR_OE_2,

	output wire                TR_DIR_3,
	output wire                TR_OE_3,

	output wire                TR_DIR_4,
	output wire                TR_OE_4,

	output wire                TR_DIR_5,
	output wire                TR_OE_5,

	input wire                 SW_USER_1,
	input wire                 SW_USER_2,

	input wire                 SLIDE_SWITCH_1,
	input wire                 SLIDE_SWITCH_2,

	output reg                 LED_1_GREEN,
	output reg                 LED_1_RED,
	output reg                 LED_1_BLUE,

	output reg                 LED_2_GREEN,
	output reg                 LED_2_RED,
	output reg                 LED_2_BLUE,

	output reg                 LED_3_GREEN,
	output reg                 LED_3_RED,
	output reg                 LED_3_BLUE,
	
    output reg                 LED_4_GREEN,
	output reg                 LED_4_RED,
	output reg                 LED_4_BLUE

	
	);

   //-----------------------------------------------
   // Parameters
   //-----------------------------------------------

   parameter                   GLOBAL_RESET_COUNT = 4'hf;
   
   //Transfer to Host State Machine
   parameter                   TRANSFER_HOST_IDLE = 1,
                               TRANSFER_HOST_LOOPBACK_START = 2,
							   TRANSFER_HOST_LOOPBACK_COMPLETE = 3;

	//Block Transfer to Host State Machine
   parameter                   BLOCK_LOOPBACK_IDLE = 1,
                               BLOCK_LOOPBACK_INIT = 2,
							   FIFO_EN_HIGH = 3,
							   FIFO_EN_LOW = 4;
							   
	//Number of GPIO Storage Registers on the MegaProLogic
   parameter                   MEGA_GPIO_MAX_NUMBER = 6;	

	//Block Transfer to Host Counter
   parameter                   BLK_TRANSFER_OUT_IDLE = 1,
			                   BLK_TRANSFER_OUT_COUNT = 2;
							   
	//Registers for Test
	wire  [3:0]                active_serial_state_out;
	wire  [11:0]                active_transfer_test_out;
	
//***************************************************************************
//* Internal Signals and Registers Declarations
//***************************************************************************
    wire                       CLK_66;
	wire                       RST_N;
	wire [1:0]                 aa;

	wire [23:0]                UC_IN;
	wire [21:0]                UC_OUT;
	
	//Trigger Signals
	reg  [7:0]                 trigger_out;
	wire [7:0]                 trigger_in_byte;	
	
    //UART_IN meta registers
	reg                        uart_meta;
	reg                        uart_mid;
	reg                        uart_txd;

	//LED registers
	reg                        led_reset;
	
	//Switch registers
	reg                        switch_reset;
	
	//Transfer registers
	reg                        transfer_out_reg;
	wire                       transfer_in_received;
	wire [7:0]                 transfer_in_byte;
	reg  [7:0]                 transfer_out_byte;
	reg  [3:0]                 transfer_to_host_counter;
	reg  [3:0]                 transfer_to_host_state;

	reg                        transfer_3_out_reg;
	wire                       transfer_3_in_received;
	reg  [7:0]                 transfer_3_out_byte;
	wire [7:0]                 transfer_3_in_byte;	


	//Block Transfer Out registers
	reg [2:0]                  state_block_loopback;
	reg                        block_out_reg;
	wire                       block_out_received;
	reg [7:0]                  block_out_counter;
	reg                        start_block_transfer;
	reg                        start_block_transfer_reg;
	wire                       block_byte_ready;
	reg [7:0]                  block_out_byte;
	reg [1:0]                  block_transfer_out_state;
	
	//Block Transfer In registers
	wire [7:0]                 block_in_data;
	wire                       block_in_rcv;
	reg                        block_in_loopback;
	wire  [7:0]                ept_length;
	wire  [7:0]                uc_length;
	wire                       block_busy;
	reg   [7:0]                block_in_data_counter;
	reg                        fifo_wren_reg;

	//Reset signals
	wire                       reset;
	reg [3:0]                  reset_counter;
	reg                        reset_signal_reg;
	
//***************************************************************************
//* 	Signal Assignments	
//***************************************************************************
   
   //Clock and Reset
   assign            aa[1] = CLK_66MHZ;
   assign            aa[0] = RST_N;
   assign            RST_N = reset_signal_reg;
    
   assign            TR_DIR_1  = 1'b1; //1 = A to B; 0 = B to A
   assign            TR_OE_1  = 1'b0;
   assign            TR_DIR_2  = 1'b1; //1 = A to B; 0 = B to A
   assign            TR_OE_2  = 1'b0;
   assign            TR_DIR_3  = 1'b1; //1 = A to B; 0 = B to A
   assign            TR_OE_3  = 1'b0;  
   assign            TR_DIR_4  = 1'b1; //1 = A to B; 0 = B to A
   assign            TR_OE_4  = 1'b0;  
   assign            TR_DIR_5  = 1'b1; //1 = A to B; 0 = B to A
   assign            TR_OE_5  = 1'b0;  

    //Block In Transfer
   //assign            block_out_byte = mem_array[block_out_counter];
    

   //**************************************************
   //* 	Reset Signal	
   //**************************************************
	always @(posedge CLK_66MHZ or negedge RST)
	begin
	  if(!RST)
	  begin
			reset_signal_reg <= 1'b0;
			reset_counter <= 0;
	  end
	  else
	  begin
		if( reset_counter < GLOBAL_RESET_COUNT )
		begin
			reset_signal_reg <= 1'b0;
			reset_counter <= reset_counter + 1'b1;
		end
		else
		begin
			reset_signal_reg <= 1'b1;
		end
	  end
	end
	
   //-----------------------------------------------
   // Register the incoming UART signal
   // 
   //-----------------------------------------------

     always @(posedge CLK_66MHZ or negedge RST_N)
  begin
     if(!RST_N)
	 begin
	     uart_meta <= 1'b0;
		 uart_mid <= 1'b0;
		 uart_txd <= 1'b0;
	end
	else
	 begin
	     uart_meta <= UART_IN;
		 uart_mid <= uart_meta;
		 uart_txd <= uart_mid;
	 end
  
  end


   //-----------------------------------------------
   // PushButton Switches 
   // Sends Trigger Out Byte to the Host
   //-----------------------------------------------

     always @(posedge CLK_66MHZ or negedge RST_N)
  begin
     if(!RST_N)
	 begin
	     trigger_out <= 0;
	end
	else
	 begin
        if(SW_USER_1)
		   trigger_out <= 8'h01;
        else
		   trigger_out <= 8'h00;
		 
        if(SW_USER_2)
		   trigger_out <= 8'h02;
        else
		   trigger_out <= 8'h00;
	 end
  
  end


   //-----------------------------------------------
   // Slide Switch Read
   //-----------------------------------------------
    always @(posedge CLK_66MHZ or negedge RST_N) 
	begin
	     if (!RST_N)
		 begin
			 transfer_3_out_reg <= 1'b0;
			 transfer_3_out_byte <= 0;
	     end 
		 else 
		 begin
			 transfer_3_out_reg <= 1'b0;
			 if(trigger_in_byte[4])	
			 begin			 
			    transfer_3_out_byte = {6'h0,SLIDE_SWITCH_2,SLIDE_SWITCH_1};
			    transfer_3_out_reg <= 1'b1;
			 end
			 else
			 begin
			    transfer_3_out_reg <= 1'b0;
			 end
	     end
    end


   //-----------------------------------------------
   // LED Set and Reset
   // 
   //-----------------------------------------------

  always @(posedge CLK_66MHZ or negedge RST_N)
  begin
    if(!RST_N)
	begin
         LED_1_GREEN = 1'bz;
         LED_1_RED = 1'bz;
         LED_1_BLUE = 1'bz;

         LED_2_GREEN = 1'bz;
         LED_2_RED = 1'bz;
         LED_2_BLUE = 1'bz;

         LED_3_GREEN = 1'bz;
         LED_3_RED = 1'bz;
         LED_3_BLUE = 1'bz;
	
         LED_4_GREEN = 1'bz;
         LED_4_RED = 1'bz;
         LED_4_BLUE = 1'bz;
    end
	else 
	begin
	     if(trigger_in_byte[0])
		 begin
		    if(transfer_3_in_byte[0])
               LED_1_GREEN = 1'b0;
			else
			   LED_1_GREEN = 1'bz;
		    if(transfer_3_in_byte[1])
               LED_1_RED = 1'b0;
			else
			   LED_1_RED = 1'bz;
		    if(transfer_3_in_byte[2])
               LED_1_BLUE = 1'b0;
			else
			   LED_1_BLUE = 1'bz;
		 end
	     if(trigger_in_byte[1])
		 begin
		    if(transfer_3_in_byte[0])
               LED_2_GREEN = 1'b0;
			else
			   LED_2_GREEN = 1'bz;
		    if(transfer_3_in_byte[1])
               LED_2_RED = 1'b0;
			else
			   LED_2_RED = 1'bz;
		    if(transfer_3_in_byte[2])
               LED_2_BLUE = 1'b0;
			else
			   LED_2_BLUE = 1'bz;
		 end
	     if(trigger_in_byte[2])
		 begin
		    if(transfer_3_in_byte[0])
               LED_3_GREEN = 1'b0;
			else
			   LED_3_GREEN = 1'bz;
		    if(transfer_3_in_byte[1])
               LED_3_RED = 1'b0;
			else
			   LED_3_RED = 1'bz;
		    if(transfer_3_in_byte[2])
               LED_3_BLUE = 1'b0;
			else
			   LED_3_BLUE = 1'bz;
		 end
	     if(trigger_in_byte[3])
		 begin
		    if(transfer_3_in_byte[0])
               LED_4_GREEN = 1'b0;
			else
			   LED_4_GREEN = 1'bz;
		    if(transfer_3_in_byte[1])
               LED_4_RED = 1'b0;
			else
			   LED_4_RED = 1'bz;
		    if(transfer_3_in_byte[2])
               LED_4_BLUE = 1'b0;
			else
			   LED_4_BLUE = 1'bz;
		 end
	end 
  end



   //-----------------------------------------------
   // Transfer to Host in Loop Back Mode 
   //-----------------------------------------------
    always @(posedge CLK_66MHZ or negedge RST_N) 
	begin
	     if (!RST_N)
		 begin
			 transfer_out_reg <= 1'b0;
			 transfer_to_host_counter <= 0;
			 transfer_to_host_state <= TRANSFER_HOST_IDLE;
	     end 
		 else 
		 begin
			     transfer_out_reg <= 1'b0;
			case(transfer_to_host_state)
             TRANSFER_HOST_IDLE:
			     if(transfer_in_received /*& transfer_in_loop_back*/)	
                 begin			 
			         transfer_out_byte = transfer_in_byte;
			         transfer_to_host_state <= TRANSFER_HOST_LOOPBACK_START;
			     end
             TRANSFER_HOST_LOOPBACK_START:
		         if(transfer_to_host_counter < 4'hf)
			     begin
		             transfer_to_host_counter <= transfer_to_host_counter + 1'd1;
			         transfer_to_host_state <= TRANSFER_HOST_LOOPBACK_START;
			     end
				 else
				 begin
			         transfer_out_reg <= 1'b1;
			         transfer_to_host_state <= TRANSFER_HOST_LOOPBACK_COMPLETE;				 
				 end
             TRANSFER_HOST_LOOPBACK_COMPLETE:
			     begin
			         transfer_out_reg <= 1'b0;
					 transfer_to_host_counter <= 0;
			         transfer_to_host_state <= TRANSFER_HOST_IDLE;
				 end
			 default:
			      begin
			          transfer_out_reg <= 1'b0;
			          transfer_to_host_counter <= 0;
			          transfer_to_host_state <= TRANSFER_HOST_IDLE;
				  end
			 endcase
	     end
    end

   //-----------------------------------------------
   // Detect Block Transfer Out start
   //-----------------------------------------------
  always @(block_out_reg or trigger_in_byte or RST_N)
  begin
     if(!RST_N)
	     start_block_transfer = 1'b0;
	 else if(block_out_reg)
	     start_block_transfer = 1'b0;
	 else if((trigger_in_byte[7]))
	     start_block_transfer = 1'b1;
  
  end
  
   //-----------------------------------------------
   // Registers to start Block Transfer Out
   //-----------------------------------------------
  always @(posedge CLK_66MHZ or negedge RST_N)
  begin
    if(!RST_N)
	begin
		 block_out_reg <= 1'b0;
		 start_block_transfer_reg <= 1'b0;
		 block_out_counter <= 0;
		 block_transfer_out_state <= BLK_TRANSFER_OUT_IDLE;
	end
	else
	begin
	     case(block_transfer_out_state)
		     BLK_TRANSFER_OUT_IDLE:
	             if(start_block_transfer )
				 begin
		             start_block_transfer_reg <= 1'b1;
		             block_transfer_out_state <= BLK_TRANSFER_OUT_COUNT;
				 end
			 BLK_TRANSFER_OUT_COUNT:
		         begin
			          block_out_reg <= 1'b1;
	                 if(block_byte_ready)
		             begin
		                 block_out_counter <= block_out_counter + 1'd1;
		             end
		             else if(block_out_counter >= MEGA_GPIO_MAX_NUMBER )
		             begin
		                 block_out_counter <= 0;
			             block_out_reg <= 1'b0;
		                 block_transfer_out_state <= BLK_TRANSFER_OUT_IDLE;
		             end
				 end
			 default:
			      begin
		                 block_out_counter <= 0;
			             block_out_reg <= 1'b0;
		                 block_transfer_out_state <= BLK_TRANSFER_OUT_IDLE;
				  end
			 endcase
	end
  end

   //-----------------------------------------------
   // Block In read fifo enable is driven by block_byte_ready
   //-----------------------------------------------
  always @(posedge CLK_66MHZ or negedge RST_N)
  begin
    if(!RST_N)
	begin
		//state_block_loopback <= BLOCK_LOOPBACK_IDLE;
	end
	else 
	begin
		     case(block_out_counter)
			 0:
			 begin
			     block_out_byte <= LB_XIOLA;
			 end
			 1:
			 begin
				 block_out_byte <= LB_XIOLC;
		     end
			 2:
		     begin
	             block_out_byte <= LB_XIOH;
		     end
			 3:
		     begin
				 block_out_byte <= LB_COMM;
			 end
			 4:
			 begin
			     block_out_byte <= LB_AD;
			 end
			 5:
			 begin
			     block_out_byte <= LB_IOH;
			 end
			 endcase
	end
  end


   //-----------------------------------------------
   // Instantiate the EPT Library
   //-----------------------------------------------

	active_serial_library	   ACTIVE_SERIAL_LIBRARY_INST
	(	
	.aa                        (aa),
	.UART_OUT                  (UART_OUT),
	.UART_IN                   (uart_txd),

	.UC_IN                     (UC_IN),
	.UC_OUT                    (UC_OUT)//,
	
	//.STATE_OUT                 (active_serial_state_out),
	//.TEST_OUT                  (active_transfer_test_out) 
	
	);

   //-----------------------------------------------
   // Instantiate the EPT Library
   //
   //
   //       Trigger In Byte 0x01 = LED1
   //       Trigger In Byte 0x02 = LED2
   //       Trigger In Byte 0x04 = LED3
   //       Trigger In Byte 0x08 = LED4
   //       Trigger In Byte 0x10 = Read Slide Switch
   //       Trigger In Byte 0x20 = Reserved
   //       Trigger In Byte 0x40 = Reserved
   //       Trigger In Byte 0x80 = Start Block Transfer of all GPIOs
   //
   //       Trigger Out Byte 0x01 = Switch1
   //       Trigger Out Byte 0x02 = Switch2
   //       Trigger Out Byte 0x04 = Reserved
   //       Trigger Out Byte 0x08 = Reserved
   //       Trigger Out Byte 0x10 = Reserved
   //       Trigger Out Byte 0x20 = Reserved
   //       Trigger Out Byte 0x40 = Reserved
   //       Trigger Out Byte 0x80 = Reserved
   //
   //
   //-----------------------------------------------
wire [22*4-1:0]  uc_out_m;
eptWireOR # (.N(4)) wireOR (UC_OUT, uc_out_m);
	active_trigger             ACTIVE_TRIGGER_INST
	(
	 .uc_clk                   (CLK_66MHZ),
	 .uc_reset                 (RST_N),
	 .uc_in                    (UC_IN),
	 .uc_out                   (uc_out_m[ 0*22 +: 22 ]),

	 .trigger_to_host          (trigger_out),
	 .trigger_to_device        (trigger_in_byte)
	
	);

	active_transfer            ACTIVE_TRANSFER_2_INST
	(
	 .uc_clk                   (CLK_66MHZ),
	 .uc_reset                 (RST_N),
	 .uc_in                    (UC_IN),
	 .uc_out                   (uc_out_m[ 1*22 +: 22 ]),
	
	 .start_transfer           (transfer_out_reg),
	 .transfer_received        (transfer_in_received),
	
	 .uc_addr                  (3'h2),

	 .transfer_to_host         (transfer_out_byte),
	 .transfer_to_device       (transfer_in_byte)	
	);
	
	active_transfer            ACTIVE_TRANSFER_3_INST
	(
	 .uc_clk                   (CLK_66MHZ),
	 .uc_reset                 (RST_N),
	 .uc_in                    (UC_IN),
	 .uc_out                   (uc_out_m[ 2*22 +: 22 ]),
	
	 .start_transfer           (transfer_3_out_reg),
	 .transfer_received        (transfer_3_in_received),
	
	 .uc_addr                  (3'h3),

	 .transfer_to_host         (transfer_3_out_byte),
	 .transfer_to_device       (transfer_3_in_byte)	
	);
	
	active_block               BLOCK_TRANSFER_INST
	(
	 .uc_clk                   (CLK_66MHZ),
	 .uc_reset                 (RST_N),
	 .uc_in                    (UC_IN),
	 .uc_out                   (uc_out_m[ 3*22 +: 22 ]),
	
	 .start_transfer           (block_out_reg),
	 .transfer_received        (block_in_rcv),
	 
	 .transfer_ready           (block_byte_ready),
	 .transfer_busy            (block_busy),

	 .ept_length               (ept_length),
	
	 .uc_addr                  (3'h4),
	 .uc_length                (MEGA_GPIO_MAX_NUMBER),

	 .transfer_to_host         (block_out_byte),
	 .transfer_to_device       (block_in_data)
	
	);

	
endmodule
